/*
 * Copyright (c) 2007 Luigi Rizzo - Universita` di Pisa
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Id: ldev_stub.c,v 1.27 2007/02/09 10:35:09 luigi Exp $
 *
 * This file implement the necessary functions to implement a
 * device driver as a loadable FreeBSD module, and on the other
 * side provides sufficient hooks to Linux device drivers to be linked
 * and operate as expected.
 *
 * compile with the following macros
 *
 * -DDEV_NAME="video"	device name (with quotes)
 * -DDRIVER_NAME=foo	driver name (without_quotes)
 * make sure ldev_devs is the array of ldev_info with vendor/prod identifiers
 * to be used for matching the device.
 */

#include <ldev_stub.h>
#include <linux/moduleparam.h>
SYSCTL_NODEn(_debug, OID_AUTO, DRIVER_NAME, CTLFLAG_RW, 0, "ldev controls");

/*
 * Fetch the unit number from the device. With devfs we cannot rely on
 * minor device numbers, but the minor() function does the job for us.
 */
#define LDEVUNIT(dev)	(minor(dev))

/*
 * Forward declarations for the driver hooks, and the cdevs structure
 * containing them. This is passed as an argument to make_dev
 */
static d_open_t  ldev_open; 
static d_close_t ldev_close;
static d_read_t  ldev_read;
static d_ioctl_t ldev_ioctl;
static d_mmap_t  ldev_mmap;
static d_poll_t  ldev_poll;
  
static struct cdevsw ldev_cdevsw = {
        .d_version      = D_VERSION,
        .d_flags        = D_NEEDGIANT,
        .d_open         = ldev_open,   
        .d_close        = ldev_close,
        .d_read         = ldev_read,
        .d_ioctl        = ldev_ioctl,
        .d_poll         = ldev_poll,
        .d_mmap         = ldev_mmap,
        .d_name         = __stringify(DRIVER_NAME),
};

/*
 * Declare the _match, _attach, _detach methods;
 * define XXX_methods() and fill with the above,
 * define XXX_devclass, XXX_driver and link with the methods.
 * Among other things, write sizeof(ldev_softc) in the driver_t
 * associated to this driver, so that the allocation of the softc
 * can be done automatically.
 */
USB_DECLARE_DRIVER(ldev);

/*
 * MODINFO_NAME(info) generates a unique name dependend on the driver's
 * name DRIVER_NAME. We use it to store device specific information from
 * the linux driver, stored directly or collected form a linker dataset
 * at the beginning of modevent_handler().
 * We need the variable to be globally visible because other source
 * files must access it, and the macro allows us to avoid name clashes
 * when the driver is statically linked in the kernel.
 * Locally, we use a static variable to avoid obfuscating the source.
 */
struct linux_dev_info MODINFO_NAME(info);
static struct linux_dev_info *linux_dev_info = &MODINFO_NAME(info);

/*
 * device lookup routine.
 * In a standard driver we would just have the following code:

    static struct ldev_info ldev_devs[] = {
	{ { vend1, prod1}, other_data  },	// one or more entries
	{ { 0, 0}  },				// terminator just in case
    };

    #define ldev_lookup(v, p, sc) 			\
	((const struct ldev_info *)usb_lookup(ldev_devs, v, sc))

 * In the linux driver, the table is sliglthy different so we
 * need to write our own loop, which is implemented in the
 * library and passed a pointer to the table.
 */

/*
 * Rather than dereferencing the usbd_*_handle fields,
 * use the usbd_get_... functions.
 * If you really cannot, you need to include these two files:

#include <machine/bus.h>
#include <dev/usb/usbdivar.h>

 */

/*
 * some debugging functions to print the status of a device
 */
static void
ldev_print_iface(usbd_interface_handle iface)
{
	usb_interface_descriptor_t *id = usbd_get_interface_descriptor(iface);
	usbd_device_handle dev;

	usbd_interface2device_handle(iface, &dev);

	printf("%s: iface %p device %p altindex %d\n",
		__FUNCTION__,
		iface, dev,
		usbd_get_interface_altindex(iface));
	printf("\tnum %d alt %d ep %d class %d int %d prot %d iInt %d\n",
		id->bInterfaceNumber,
		id->bAlternateSetting,
		id->bNumEndpoints,
		id->bInterfaceClass,
		id->bInterfaceSubClass,
		id->bInterfaceProtocol,
		id->iInterface);
}

static void
ldev_print_device(usbd_device_handle udev)
{
	usb_device_descriptor_t *udd = usbd_get_device_descriptor(udev);
	// usb_config_descriptor_t *ucd = udev->cdesc;

	int nc = udd->bNumConfigurations;
	printf("%s: dev %p configs %d\n",
		__FUNCTION__, udev,
		nc);
}

static void
ldev_print_info(struct usb_attach_arg *uaa)
{
	int ni;

	printf("%s: 0x%04x: 0x%04x port %d config %d iface %d nifaces %d\n",
		__FUNCTION__,
		uaa->vendor, uaa->product,
		uaa->port, uaa->configno, uaa->ifaceno, uaa->nifaces);
	ldev_print_device(uaa->device);
	printf("\tdevice %p ifaces %p\n",
		uaa->device, uaa->ifaces);
	for (ni = 0; ni < uaa->nifaces; ni++) {
		printf("\tthis if %p [%d] %p\n",
			uaa->iface, ni, uaa->ifaces[ni]);
	}
	if (uaa->ifaceno >= 0 && uaa->iface != NULL)
		ldev_print_iface(uaa->iface);
}


static int
ldev_match(device_t dev)
{
	// struct ldev_softc *sc = device_get_softc(dev);
        struct usb_attach_arg *uaa = device_get_ivars(dev);
        usb_interface_descriptor_t *id;

	if (0)
	ldev_print_info(uaa);
	if (ldev_lookup(dev, linux_dev_info->device_table) == NULL)
		return UMATCH_NONE;

	/*
	 * Driver loaded when device was already plugged in,
	 * we have to claim all interfaces or get none...
	 */
        if (uaa->usegeneric)
		return  UMATCH_VENDOR_PRODUCT;

        /* We don't want all interfaces */
        if (uaa->iface == NULL)
                return UMATCH_NONE;

        id = usbd_get_interface_descriptor(uaa->iface);
        if (id == NULL) {
                printf("ldev: failed to get interface descriptor\n");
                return UMATCH_NONE;
        }

        if (id->bInterfaceNumber != 0)
                return UMATCH_NONE;
        return  UMATCH_VENDOR_PRODUCT;
}

static int
ldev_attach(device_t dev)
{
        struct ldev_softc *sc = device_get_softc(dev);
        struct usb_attach_arg *uaa = device_get_ivars(dev);
        char devinfo[1024];
        const char *tmpstr;
        const struct usb_device_id *info;
        int unit = device_get_unit(dev);
	const struct usb_driver *dp;
  
        /* set up udev early so the lookup/attach routine can do more */
        sc->udev = uaa->device;
	sc->sc_iface = uaa->iface;
        sc->sc_dev = dev;
 
printf("%s: sc at %p, l_u_d at %p\n", __FUNCTION__, sc, &sc->l_u_d);

	info = ldev_lookup(dev, linux_dev_info->device_table);
        if (info == NULL) {
                device_printf(dev,
            "attach error vendor/product mismatch (vendor=0x%x product=0x%x)\n",
                        uaa->vendor,uaa->product);
                return ENXIO;
        }

	 if(uaa->iface == NULL) {
		device_printf(dev, "attach: iface is NULL\n");
		return ENXIO;
	}
		
	/* prepare the linux emulation layer */
	if (fill_linux_descriptors(dev, &sc->l_u_d)) {
		device_printf(dev, "attach: failed to allocate linux desc\n");
		return ENOMEM;
	}

	/* prepare to call the linux routine */
	dp = linux_dev_info->usb_driver_desc;
	if (dp == NULL || dp->probe == NULL) {
		device_printf(dev, "linux attach missing, sorry\n");
		return ENXIO;
	}

	if (dp->probe(&sc->l_u_d.intf, info) != 0) {
		device_printf(dev, "linux attach failed, sorry\n");
		return ENXIO;
	}

        usbd_devinfo(uaa->device, 0, devinfo);
        device_set_desc_copy(dev, devinfo);
        device_printf(dev, "%s\n", devinfo);

        mtx_init(&sc->ptrlock,device_get_name(sc->sc_dev),NULL,MTX_DEF);

        resource_string_value(__stringify(DRIVER_NAME), unit, "devname", &tmpstr);
        sc->sc_dev_t = make_dev(&ldev_cdevsw, unit, UID_ROOT, GID_OPERATOR,
		0666, "%s%d", DEV_NAME, unit);
	/* call the attach routine (where ???) */
	usbd_add_drv_event(USB_EVENT_DRIVER_ATTACH, sc->udev,USBDEV(sc->sc_dev));
	return 0; /* success */
}

static int
ldev_detach(device_t dev)
{
        struct ldev_softc *sc = device_get_softc(dev);

        sc->error_status = EPIPE;	/* record detaching */
	/* wait for userland to really close */
	while (sc->vopen) {
                if(sc->state & LDEV_ASLEEP)
                        wakeup(sc);
                if(sc->state & LDEV_POLL) {
                        sc->state &= ~LDEV_POLL;
                        selwakeuppri(&sc->rsel,PZERO);
                }
                device_printf(sc->sc_dev, "Disconnected while device is in use!\n");
                usb_detach_wait(USBDEV(sc->sc_dev));
        }
	/*  free resources etc. */
	if (sc->l_u_d.mem)	/* memory allocated for linux emulation */
		free(sc->l_u_d.mem, M_USBDEV);
        if(sc->sc_dev_t != NULL)
                destroy_dev(sc->sc_dev_t);

        mtx_destroy(&sc->ptrlock);
        /* free memory ? */
        usbd_add_drv_event(USB_EVENT_DRIVER_DETACH, sc->udev,USBDEV(sc->sc_dev));
        return 0;
}

static int
ldev_open(struct cdev *dev, int flag, int mode, usb_proc_ptr p)
{
        int err = 0;
        int unit = LDEVUNIT(dev);
        struct ldev_softc *sc = devclass_get_softc(ldev_devclass, unit);
  
        if (sc == NULL)
                return ENXIO;
        
        if (sc->error_status == EPIPE)	/* detach in progress */
                return sc->error_status;

	/* if vopen is 0, set to 1 and return success */
	if (atomic_cmpset_int(&sc->vopen, 0, 1) == 0)	/* already open */
                return EBUSY;
 
	err = -sc->l_u_d.fops->open((struct inode *)sc, &sc->l_u_d.file);
	if (err)
		goto bad;

	/* FILL HERE: fire the device */
	return 0;

bad:
	/* free memory */
	sc->vopen = 0;
	return err;
}

/* Watch out for races between close and detach */
static int
ldev_close(struct cdev *dev, int flag, int mode, usb_proc_ptr p)
{
        int unit = LDEVUNIT(dev);
	struct ldev_softc *sc = devclass_get_softc(ldev_devclass, unit);
	int err;

	if (sc == NULL)
		return ENXIO;

	err = -sc->l_u_d.fops->release((struct inode *)sc, &sc->l_u_d.file);
	sc->vopen = 0;

	usb_detach_wakeup(USBDEV(sc->sc_dev));	/* if a detach is pending */
	return err;
}

static int
ldev_read(struct cdev *dev, struct uio *uio, int flag)
{
        int unit = LDEVUNIT(dev);
        struct ldev_softc *sc = devclass_get_softc(ldev_devclass, unit);
        int count = uio->uio_resid;
        int err;

        if (sc->error_status)
                return sc->error_status;

	/*
	 * call the linux handler, using 'buf' as an alias to the
	 * struct uio, which in the end is passed to copy_to_user.
	 * Note, the linux routine returns the number of bytes read.
	 */
	err = sc->l_u_d.fops->read(&sc->l_u_d.file, (char *)uio, count, NULL);
#if 0
        if(1 /* sc->image_read_pos == 0 */) {
		while ((err = wait_data(sc)) == EBUSY) {
                       if(flag & O_NONBLOCK)
                                return EWOULDBLOCK;
                        sc->state |= LDEV_ASLEEP;
                        err = tsleep(sc, PZERO | PCATCH, "ldevrd", 0);
                        sc->state &= ~LDEV_ASLEEP;

                        if(sc->error_status)
                                return sc->error_status;
                        else if(err)
                                return err;
                }
                if(err)
                        return err;
        }

        err = uiomove(sc->data_src, count, uio);
#endif
	return 0;	/* no errors for the time being... */
}

/*
 * This should not be here, but on the other hand we are only
 * doing webcams now... we want to know the ioctls used for them
 * in order to filter out some.
 */
#include <sys/ioccom.h>
#include <linux/videodev.h>

static int
ldev_ioctl(struct cdev *dev, u_long cmd, caddr_t addr, int flag, usb_proc_ptr p)
{
        int unit = LDEVUNIT(dev);
        struct ldev_softc *sc = devclass_get_softc(ldev_devclass, unit);
	int err;

        if (sc->error_status)
                return sc->error_status;

	printf("%s: called\n", __FUNCTION__);
	/* call the device-specific ioctl handler */

	/*
	 * filter out some non-implemented commands
	 */
	switch (cmd) {
	case VIDIOCGMBUF:
	case VIDIOCMCAPTURE:
	case VIDIOCSYNC:
		/* mmap not implemented yet, so fail */
		return EINVAL;
	}
	err = -sc->l_u_d.fops->ioctl((struct inode *)sc,
		&sc->l_u_d.file, cmd, (unsigned long)addr);
        return err;
}

/* XXX this is still incomplete */
static int
ldev_poll(struct cdev *dev, int events, usb_proc_ptr p)
{
        int unit = LDEVUNIT(dev);
        struct ldev_softc *sc = devclass_get_softc(ldev_devclass, unit);
        int revents = 0;

	printf("%s: called\n", __FUNCTION__);
        if(sc->error_status)
                return sc->error_status;

        if (events & (POLLIN | POLLRDNORM)) {
                mtx_lock(&sc->ptrlock);
                if(sc->full_frames == NULL) {
                        sc->state |= LDEV_POLL;
                        selrecord(p,&sc->rsel);
                } else {
                        revents |= events & (POLLIN | POLLRDNORM);
                }
                mtx_unlock(&sc->ptrlock);
        }
        if (events & (POLLOUT | POLLWRNORM)) {
                /* write is not allowed, so no point blocking on it */
                revents |= events & (POLLOUT | POLLWRNORM);
        }
        return revents;
}

static int
ldev_mmap(struct cdev *dev, vm_offset_t offset, vm_paddr_t *paddr, int nprot)
{
#ifdef USE_MMAP
        int unit = LDEVUNIT(dev);
        struct ldev_softc *sc = devclass_get_softc(ldev_devclass, unit);

        if (sc == NULL)
                return ENXIO;

        if (sc->error_status)
                return sc->error_status;

        if(nprot & PROT_EXEC || sc->image_data == NULL || offset < 0 || offset >= sc->pwc_mbufs * round_page(sc->len_per_image))
                return -1;

        *paddr = vtophys(((char*)sc->image_data) + offset);
        return 0;
#else
        return ENXIO;
#endif
}

static int modevent_handler(module_t mod, int what, void *arg)
{
	struct linux_dev_info *i = linux_dev_info;

	linux_devinfo_init(i);

	switch (what) {
	case MOD_LOAD:
		if (i->init_f)
			i->init_f();
		break;

	case MOD_UNLOAD:
		if (i->exit_f)
			i->exit_f();
		break;

	case MOD_QUIESCE:	/* called before module_unload */
		return 0;
	}
	return 0;
}

/* 
 * we need to call tsleep, which in the linux emulation is a #define,
 * so we call it from a wrapper.
 * XXX todo: make wc have a reference to
 */
int
wait_event_body(void *wc, int flags)
{
        int err;
        if (flags & O_NONBLOCK)
                return EWOULDBLOCK; 
  
        // sc->state |= PWC_ASLEEP;
	if (0)
		DBG("calling tsleep on %p\n", wc);
        err = tsleep(wc, PZERO | PCATCH, "lmod", 0);
        //sc->state &= ~PWC_ASLEEP;
        return err;
}  

#define	__A(x, a, b, c, d, e)	DRIVER_MODULE(x	, a, b,c,d,e)
__A(DRIVER_NAME, uhub, ldev_driver, ldev_devclass, modevent_handler, 0);
// DRIVER_MODULE(DRIVER_NAME, uhub, ldev_driver, ldev_devclass, modevent_handler, 0);
